/*!
 * vue-form-container v1.0.0
 * (c) 2018-present Vitor Luiz Cavalcanti <vitorluizc@outlook.com> (https://vitorluizc.github.io)
 * Released under the MIT License.
 */
(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global.VueFormContainer = factory());
}(this, (function () { 'use strict';

  /*!
   * isobject <https://github.com/jonschlinkert/isobject>
   *
   * Copyright (c) 2014-2017, Jon Schlinkert.
   * Released under the MIT License.
   */

  var isobject = function isObject(val) {
    return val != null && typeof val === 'object' && Array.isArray(val) === false;
  };

  /*!
   * get-value <https://github.com/jonschlinkert/get-value>
   *
   * Copyright (c) 2014-2018, Jon Schlinkert.
   * Released under the MIT License.
   */



  var getValue = function(target, path, options) {
    if (!isobject(options)) {
      options = { default: options };
    }

    if (!isValidObject(target)) {
      return typeof options.default !== 'undefined' ? options.default : target;
    }

    if (typeof path === 'number') {
      path = String(path);
    }

    const isArray = Array.isArray(path);
    const isString = typeof path === 'string';
    const splitChar = options.separator || '.';
    const joinChar = options.joinChar || (typeof splitChar === 'string' ? splitChar : '.');

    if (!isString && !isArray) {
      return target;
    }

    if (isString && path in target) {
      return isValid(path, target, options) ? target[path] : options.default;
    }

    let segs = isArray ? path : split(path, splitChar, options);
    let len = segs.length;
    let idx = 0;

    do {
      let prop = segs[idx];
      if (typeof prop === 'number') {
        prop = String(prop);
      }

      while (prop && prop.slice(-1) === '\\') {
        prop = join([prop.slice(0, -1), segs[++idx] || ''], joinChar, options);
      }

      if (prop in target) {
        if (!isValid(prop, target, options)) {
          return options.default;
        }

        target = target[prop];
      } else {
        let hasProp = false;
        let n = idx + 1;

        while (n < len) {
          prop = join([prop, segs[n++]], joinChar, options);

          if ((hasProp = prop in target)) {
            if (!isValid(prop, target, options)) {
              return options.default;
            }

            target = target[prop];
            idx = n - 1;
            break;
          }
        }

        if (!hasProp) {
          return options.default;
        }
      }
    } while (++idx < len && isValidObject(target));

    if (idx === len) {
      return target;
    }

    return options.default;
  };

  function join(segs, joinChar, options) {
    if (typeof options.join === 'function') {
      return options.join(segs);
    }
    return segs[0] + joinChar + segs[1];
  }

  function split(path, splitChar, options) {
    if (typeof options.split === 'function') {
      return options.split(path);
    }
    return path.split(splitChar);
  }

  function isValid(key, target, options) {
    if (typeof options.isValid === 'function') {
      return options.isValid(key, target);
    }
    return true;
  }

  function isValidObject(val) {
    return isobject(val) || Array.isArray(val) || typeof val === 'function';
  }

  /*!
   * valite v0.7.0
   * (c) 2018-present Vitor Luiz Cavalcanti <vitorluizc@outlook.com> (https://vitorluizc.github.io)
   * Released under the MIT License.
   */

  /**
   * A valite's validator error class.
   */
  var ValidatorError = /*@__PURE__*/(function (Error) {
    function ValidatorError(message) {
      Error.call(this, message);
      this.name = 'ValidatorError';
      this.message = message;
    }

    if ( Error ) ValidatorError.__proto__ = Error;
    ValidatorError.prototype = Object.create( Error && Error.prototype );
    ValidatorError.prototype.constructor = ValidatorError;

    return ValidatorError;
  }(Error));

  /**
   * Check if value is a message.
   * @param value
   */

  var isMessage = function (value) {
    if (typeof value !== 'string' && value !== true) { throw new ValidatorError('Validator should return `true` or a message instead.'); }
    if (typeof value === 'string' && !value.trim()) { throw new ValidatorError('Validator message shouldn\'t be empty.'); }
    return value !== true;
  };
  /**
   * Check if value is a message schema.
   * @param value
   */

  var isMessageSchema = function (value) { return value && typeof value === 'object'; };

  /**
   * Get an Array of keys from a given object.
   */

  var keys = Object.keys;
  /**
   * Sequentially executes validators over the value. Returns the first error
   * message or `null` otherwise.
   * @param value
   * @param validators
   */

  var validate = function (value, ref) {
    if ( ref === void 0 ) ref = [];
    var validator = ref[0];
    var validators = ref.slice(1);

    return typeof validator !== 'function' ? Promise.resolve(null) : Promise.resolve(validator(value)).then(function (message) { return isMessage(message) ? message : validate(value, validators); });
  };
  /**
   * Concurrenly validates an object with an schema. It executes validators for
   * deep properties and returns a message schema.
   * @param object
   * @param schema
   */

  var validateObject = function (object, schema) {
    var errors = Object.create(null);
    var resolutions = keys(schema).map(function (key) { return validate(getValue(object, key), schema[key]).then(function (message) {
      errors[key] = message;
    }); });
    return Promise.all(resolutions).then(function () { return errors; });
  };
  /**
   * Check if error message or error message schema is empty, and therefore valid.
   * @param error
   */

  var isValid$1 = function (error) { return !isMessageSchema(error) ? error === null : keys(error).every(function (key) { return error[key] === null; }); };

  /**
   * Defines acessors (get and set) to key on object.
   * @param {object} object
   * @param {string} key
   * @param {{ get: () => any, set: (value: any) => void }} acessors
   */
  var definePropertyAcessors = function (object, key, ref) {
    var get = ref.get;
    var set = ref.set;

    return Object.defineProperty(object, key, {
    get: get,
    set: set,
    enumerable: true,
    configurable: false
  });
  };
  /**
   * Defines a read-only property on target object.
   * @param {Object.<*, *>} target
   * @param {string} key
   * @param {function():any} get
   */

  var defineReadOnlyProperty = function (target, key, get) { return definePropertyAcessors(target, key, {
    get: get,
    set: function () {
      throw new Error(("Cannot assign to readonly property \"" + key + "\"."));
    }
  }); };

  /**
   * A function to check if value attends to predicate.
   * @callback Predicate
   * @param {*} value Any value could be checked.
   * @returns {boolean} `true` if value attends to predicate.
   */

  /**
   * Value returned by `typeof` to determine what is value's type.
   * @typedef {('number' | 'object' | 'string' | 'symbol' | 'boolean' | 'function' | 'undefined')} Type
   */

  /**
   * Check if value is of type.
   * @param {Type} type
   * @returns {Predicate}
   */
  var isType = function (type) { return function (value) { return typeof value === type; }; };
  /**
   * Check if value is an object.
   * @param {*} value
   * @returns {boolean}
   */

  var isObject = function (value) { return !!value && isType('object')(value); };
  /**
   * Check if value is an Array and every element attends the predicate.
   * @param {Predicate} predicate
   * @returns {Predicate}
   */

  var isEvery = function (predicate) { return function (value) { return Array.isArray(value) && value.every(predicate); }; };
  /**
   * Check if every object properties attends the predicate.
   * @param {Predicate} predicate
   * @returns {Predicate}
   */

  var isEveryProperty = function (predicate) { return function (value) {
    if (!isObject(value)) { return false; }
    var properties = Object.keys(value).map(function (key) { return value[key]; });
    return properties.every(isEvery(predicate));
  }; };

  /**
   * Model of global form accessor object.
   * @typedef {Object} FormAcessor
   * @property {Object.<string, *>} fields
   * @property {Object.<string, string>} errors
   * @property {function(string,*):void} update
   * @property {boolean} isValid
   * @property {boolean} isLoading
   * @property {function():Promise.<void>} validateForm
   * @property {function(string):Promise.<void>} validateField
   */

  var script = {
    props: {
      schema: {
        type: Object,
        required: true,
        validator: isEveryProperty(isType('function'))
      },
      initial: {
        type: Object,
        default: function () { return Object.create(null); },
        validator: isObject
      }
    },

    data: function data() {
      return {
        ticks: 0,
        errors: Object.create(null),
        values: Object.create(null)
      };
    },

    watch: {
      /**
       * When schema is changed it updates values with it.
       * @param {object} schema
       */
      schema: function schema(schema$1) {
        this.setupValuesWith(schema$1, this.values);
      }

    },
    computed: {
      fields: function fields() {
        var this$1 = this;

        var fields = Object.create(null);
        Object.keys(this.values).forEach(function (field) { return definePropertyAcessors(fields, field, {
          get: function () { return this$1.values[field]; },
          set: function (value) { return this$1.update(field, value); }
        }); });
        return fields;
      },

      scope: function scope() {
        var this$1 = this;

        var scope = {
          update: this.update,
          validateForm: this.validateForm,
          validateField: this.validateField
        };
        defineReadOnlyProperty(scope, 'fields', function () { return this$1.fields; });
        defineReadOnlyProperty(scope, 'errors', function () { return this$1.errors; });
        defineReadOnlyProperty(scope, 'isValid', function () { return this$1.isValid; });
        defineReadOnlyProperty(scope, 'isLoading', function () { return this$1.isLoading; });
        return scope;
      },

      isLoading: function isLoading() {
        return this.ticks > 0;
      },

      isValid: function isValid$1$$1() {
        return !this.isLoading && isValid$1(this.errors);
      }

    },
    methods: {
      /**
       * Setups state using an schema and value objects.
       * @param {object} schema
       * @param {object} values
       */
      setupValuesWith: function setupValuesWith(schema, values) {
        var keys = Object.keys(Object.assign({}, schema, values));
        this.values = keys.reduce(function (state, key) {
          state[key] = values[key] || undefined;
          return state;
        }, Object.create(null));
      },

      /**
       * Updates a field and trigger its validation.
       * @param {string} field Field's name.
       * @param {*} value Field's value.
       */
      update: function update(field, value) {
        this.$set(this.values, field, value);
        this.validateField(field);
      },

      /**
       * Validates a field and updates Form properties.
       * @param {string} field Field's name.
       * @returns {Promise.<void>}
       */
      validateField: function validateField(field) {
        var this$1 = this;

        this.ticks += 1;
        return validate(this.values[field], this.schema[field]).then(function (error) {
          this$1.$set(this$1.errors, field, error);
          this$1.ticks -= 1;
        }).catch(function (_) {
          console.error(("Can't validate field \"" + field + "\"."));
          this$1.ticks -= 1;
        });
      },

      /**
       * Validates all fields and updates Form properties.
       * @returns {Promise.<void>}
       */
      validateForm: function validateForm() {
        var this$1 = this;

        this.ticks += 1;
        return validateObject(this.values, this.schema).then(function (errors) {
          this$1.errors = errors;
          this$1.ticks -= 1;
        }).catch(function (_) {
          console.error('Can\'t validate form.');
          this$1.ticks -= 1;
        });
      }

    },

    render: function render() {
      return this.$scopedSlots.default(this.scope);
    },

    mounted: function mounted() {
      this.setupValuesWith(this.schema, this.initial);
    }

  };

  /* script */
              var __vue_script__ = script;
              
  /* template */

    /* style */
    var __vue_inject_styles__ = undefined;
    /* scoped */
    var __vue_scope_id__ = undefined;
    /* module identifier */
    var __vue_module_identifier__ = undefined;
    /* functional template */
    var __vue_is_functional_template__ = undefined;
    /* component normalizer */
    function __vue_normalize__(
      template, style, script$$1,
      scope, functional, moduleIdentifier,
      createInjector, createInjectorSSR
    ) {
      var component = (typeof script$$1 === 'function' ? script$$1.options : script$$1) || {};

      // For security concerns, we use only base name in production mode.
      component.__file = "FormContainer.vue";

      if (!component.render) {
        component.render = template.render;
        component.staticRenderFns = template.staticRenderFns;
        component._compiled = true;

        if (functional) { component.functional = true; }
      }

      component._scopeId = scope;

      return component
    }
    /* style inject */
    
    /* style inject SSR */
    

    
    var FormContainer = __vue_normalize__(
      {},
      __vue_inject_styles__,
      __vue_script__,
      __vue_scope_id__,
      __vue_is_functional_template__,
      __vue_module_identifier__,
      undefined,
      undefined
    );

  return FormContainer;

})));
